using SoundFlow.Backends.MiniAudio;
using SoundFlow.Components;
using SoundFlow.Enums;
using SoundFlow.Providers;

using System.Globalization;
using System.IO.Ports;
using System.Numerics;
using System.Text.Json;
using System.Timers;

namespace PC3Player
{
    public partial class MainForm : Form
    {
        static PlaylistConfig Config = new();

        static SerialPort ControllerPort;
        static SerialPort ShowSelectorPort;

        static Random SelectRandom = new Random();

        System.Timers.Timer FrameTimer;
        System.Timers.Timer ResyncTimer;
        System.Timers.Timer GUIUpdateTimer;

        SoundPlayer AudioPlayer;
        StreamDataProvider dataProvider;

        static int ShowtapeIndex = 0;
        static int PlaylistIndex = 0;
        static int IntermissionSecondsLeft = 0;
        static int IntermissionTotalTime = 0;
        static int ControllerBits;
        static int SelectorButtons;

        static bool ControllerConnected = false;
        static bool ShowSelectorConnected = false;

        public static bool ChannelMappingLoaded = false;

        static bool Playing = false;
        static bool Paused = false;
        static bool InvalidateSave = false;
        static bool Intermission;

        static string[] ShowtapeFrames;
        static PlaylistShow CurrentlyPlaying;

        public MainForm()
        {
            InitializeComponent();
            GUIUpdateTimer = new System.Timers.Timer(1000);
            GUIUpdateTimer.Elapsed += GUIUpdateTick;
            GUIUpdateTimer.AutoReset = true;
            MiniAudioEngine audioEngine = new MiniAudioEngine(48000, Capability.Playback);
        }

        private void UpdateListView()
        {
            PlaylistListView.Items.Clear();
            foreach (PlaylistShow s in Config.Shows) PlaylistListView.Items.Add(new ListViewItem(new[] { s.Name, s.Length }));
        }

        private void LoadChannelMapButton_Click(object sender, EventArgs e)
        {
            if (!ControllerConnected)
            {
                MessageBox.Show("You must connect a Controller first.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            if (SelectMappingFileDialog.ShowDialog() == DialogResult.OK)
            {
                string tempMappingData = File.ReadAllText(SelectMappingFileDialog.FileName);
                if (!tempMappingData.StartsWith("PC3MAPPING,2,"))
                {
                    MessageBox.Show("Selected mapping file is not a Version 2 PinkConnection channel map.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                string[] header = tempMappingData.Split(";")[0].Split(",");

                List<int> targetBits = new List<int>();
                foreach (string s in tempMappingData.Split(";")[1].Split(","))
                {
                    targetBits.Add(int.Parse(s));
                }

                if (ChannelMappingLoaded)
                {
                    if (header[3] != Config.Mapping.Type)
                    {
                        MessageBox.Show(
                            $"Selected Channel Map \"{header[2]}\"'s type does not match the old Channel Map's type." +
                            $"\nOld Channel Map Type: {Config.Mapping.Type}" +
                            $"\nNew Channel Map Type: {header[3]}"
                        , "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }
                }

                if (targetBits.Count != ControllerBits)
                {
                    MessageBox.Show("The mapped channel count is not equal to the connected controller's bit count.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                if (targetBits.Count % 4 != 0)
                {
                    MessageBox.Show("The mapped channel count is not divisible by 4.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                List<int> invertedBits = new List<int>();
                foreach (string s in tempMappingData.Split(";")[2].Split(","))
                {
                    int ints = int.Parse(s);
                    if (ints == 0) continue;
                    invertedBits.Add(ints);
                }

                Config.Mapping = new ChannelMapping(header[2], header[3], targetBits.ToArray(), invertedBits.ToArray());
                ChannelMappingLoaded = true;

                ChannelMappingGroupBox.Text = $"Channel Map: {Config.Mapping.Name}";
                InvalidateSave = true;
            }
        }

        private void ControllerAutoDetectButton_Click(object sender, EventArgs e)
        {
            if (ControllerConnected)
            {
                MessageBox.Show("A Controller is already connected.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            string successPortName = "";

            foreach (string portName in SerialPort.GetPortNames())
            {
                if (ControllerConnected) break;
                try
                {
                    SerialPort tempPort = new SerialPort(portName, 9600, Parity.None, 8, StopBits.One);
                    tempPort.Open();

                    for (int i = 0; i < 10; i++)
                    {
                        tempPort.Write("! ");
                        Thread.Sleep(100);
                        string readAttempt = tempPort.ReadExisting();
                        if (readAttempt.Split(",")[0] == "PC3")
                        {
                            ControllerBits = int.Parse(readAttempt.Split(",")[1]);
                            ControllerConnected = true;
                            tempPort.Close();
                            ControllerPort = new SerialPort(portName, 9600, Parity.None, 8, StopBits.One);
                            ControllerPort.Open();
                            successPortName = portName;
                            break;
                        }
                    }
                    tempPort.Close();
                }
                catch (Exception) { continue; }
            }
            if (!ControllerConnected)
            {
                MessageBox.Show("Could not detect a PinkConnection3 controller on any of your serial ports.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            ControllerGroupBox.Text = $"Controller: {ControllerBits} channel on {successPortName}";
            ControllerAutoDetectButton.Enabled = false;
            ControllerSelectSerialPortButton.Enabled = false;
        }

        private void ControllerSelectSerialPortButton_Click(object sender, EventArgs e)
        {
            if (ControllerConnected)
            {
                MessageBox.Show("A Controller is already connected.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            SerialPortSelector portDialog = new SerialPortSelector();
            if (portDialog.ShowDialog() != DialogResult.OK) return;

            string successPortName = "";

            try
            {
                SerialPort tempPort = new SerialPort(portDialog.PortName, 9600, Parity.None, 8, StopBits.One);
                tempPort.Open();

                for (int i = 0; i < 10; i++)
                {
                    tempPort.Write("! ");
                    Thread.Sleep(100);
                    string readAttempt = tempPort.ReadExisting();
                    if (readAttempt.Split(",")[0] == "PC3")
                    {
                        ControllerBits = int.Parse(readAttempt.Split(",")[1]);
                        ControllerConnected = true;
                        tempPort.Close();
                        ControllerPort = new SerialPort(portDialog.PortName, 9600, Parity.None, 8, StopBits.One);
                        ControllerPort.Open();
                        successPortName = portDialog.PortName;
                        break;
                    }
                }

                tempPort.Close();

                if (!ControllerConnected)
                {
                    MessageBox.Show("Could not detect a PinkConnection3 controller on the specified serial port.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

            }
            catch (Exception ex)
            {
                MessageBox.Show("Could not detect a PinkConnection3 controller on the specified serial port.\n" + ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            ControllerGroupBox.Text = $"Controller: {ControllerBits} channel on {successPortName}";
            ControllerAutoDetectButton.Enabled = false;
            ControllerSelectSerialPortButton.Enabled = false;
        }

        private void AddShowButton_Click(object sender, EventArgs e)
        {
            if (!ChannelMappingLoaded)
            {
                MessageBox.Show("You must load a channel mapping first.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            if (SelectShowFileDialog.ShowDialog() == DialogResult.OK)
            {
                foreach (string showFilePath in SelectShowFileDialog.FileNames)
                {
                    string tempUSTData = File.ReadAllText(showFilePath);
                    if (!tempUSTData.StartsWith("UST,2,"))
                    {
                        MessageBox.Show($"Selected showtape \"{Path.GetFileNameWithoutExtension(showFilePath)}\" is not a UST version 2 showtape.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }

                    string[] headerData = tempUSTData.Split(';')[0].Split(',');
                    string[] stringyBits = tempUSTData.Split(';')[1].Split(',');

                    TimeSpan time = TimeSpan.FromSeconds(stringyBits.Length / 60);
                    string formattedLength = time.ToString(@"hh\:mm\:ss");

                    tempUSTData = null;
                    stringyBits = null;

                    if (headerData[3] != Config.Mapping.Type)
                    {
                        MessageBox.Show(
                            $"Selected Show \"{headerData[2]}\"'s type does not match the loaded Channel Map." +
                            $"\nLoaded Channel Map Type: {Config.Mapping.Type}" +
                            $"\nShow Type: {headerData[3]}"
                        , "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }

                    Config.Shows.Add(new PlaylistShow(headerData[2], formattedLength, headerData[3], showFilePath));
                }
                UpdateListView();
                InvalidateSave = true;
            }
        }

        private void RemoveButton_Click(object sender, EventArgs e)
        {
            if (PlaylistListView.SelectedIndices.Count == 0) return;
            Config = RemoveShowAt(PlaylistListView.SelectedIndices[0], Config);
            UpdateListView();
            InvalidateSave = true;
        }

        private void MoveUpButton_Click(object sender, EventArgs e)
        {
            if (PlaylistListView.SelectedIndices.Count == 0) return;
            int selectedIndex = PlaylistListView.SelectedIndices[0];
            if (selectedIndex == 0) return;
            int assindex = 0;
            foreach (int? assignment in Config.SelectorAssignments)
            {
                if (assignment == null) continue;
                if (assignment == selectedIndex) Config.SelectorAssignments[assindex]--;
                if (assignment == selectedIndex - 1) Config.SelectorAssignments[assindex]++;
                assindex++;
            }
            PlaylistShow temp = Config.Shows[selectedIndex];
            Config.Shows.RemoveAt(selectedIndex);
            Config.Shows.Insert(selectedIndex - 1, temp);
            UpdateListView();
            InvalidateSave = true;
        }

        private void MoveDownButton_Click(object sender, EventArgs e)
        {
            if (PlaylistListView.SelectedIndices.Count == 0) return;
            int selectedIndex = PlaylistListView.SelectedIndices[0];
            if (selectedIndex >= PlaylistListView.Items.Count - 1) return;
            int assindex = 0;
            foreach (int? assignment in Config.SelectorAssignments)
            {
                if (assignment == null) continue;
                if (assignment == selectedIndex) Config.SelectorAssignments[assindex]++;
                if (assignment == selectedIndex + 1) Config.SelectorAssignments[assindex]--;
                assindex++;
            }
            PlaylistShow temp = Config.Shows[selectedIndex];
            Config.Shows.RemoveAt(selectedIndex);
            Config.Shows.Insert(selectedIndex + 1, temp);
            UpdateListView();
            InvalidateSave = true;
        }

        private void AdvancedSettingsButton_Click(object sender, EventArgs e)
        {
            AdvancedSettingsForm advancedSettingsForm = new AdvancedSettingsForm(Config);
            if (Config != advancedSettingsForm.EditingObject) InvalidateSave = true;
            if (advancedSettingsForm.ShowDialog() == DialogResult.OK) Config = advancedSettingsForm.EditingObject;
        }

        private void ShowSelectorAutoDetectButton_Click(object sender, EventArgs e)
        {
            if (ShowSelectorConnected)
            {
                MessageBox.Show("A Show Selector is already connected.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            string successPortName = "";

            foreach (string portName in SerialPort.GetPortNames())
            {
                if (ShowSelectorConnected) break;
                try
                {
                    SerialPort tempPort = new SerialPort(portName, 9600, Parity.None, 8, StopBits.One);
                    tempPort.Open();

                    for (int i = 0; i < 10; i++)
                    {
                        tempPort.Write("\" ");
                        Thread.Sleep(100);
                        string readAttempt = tempPort.ReadExisting();
                        if (readAttempt.Split(",")[0] == "PCSELECTOR")
                        {
                            SelectorButtons = int.Parse(readAttempt.Split(",")[1]);
                            ShowSelectorConnected = true;
                            tempPort.Close();
                            ShowSelectorPort = new SerialPort(portName, 9600, Parity.None, 8, StopBits.One);
                            ShowSelectorPort.Open();
                            successPortName = portName;
                            break;
                        }
                    }
                    tempPort.Close();
                }
                catch (Exception) { continue; }
            }
            if (!ShowSelectorConnected)
            {
                MessageBox.Show("Could not detect a PinkConnection Show Selector on any of your serial ports.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            ShowSelectorGroupBox.Text = $"{SelectorButtons} button on {successPortName}";
            ShowSelectorAutoDetectButton.Enabled = false;
            ShowSelectorSelectPortButton.Enabled = false;
        }

        private void ShowSelectorSelectPortButton_Click(object sender, EventArgs e)
        {
            if (ShowSelectorConnected)
            {
                MessageBox.Show("A Show Selector is already connected.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            SerialPortSelector portDialog = new SerialPortSelector();
            if (portDialog.ShowDialog() != DialogResult.OK) return;

            string successPortName = "";

            try
            {
                SerialPort tempPort = new SerialPort(portDialog.PortName, 9600, Parity.None, 8, StopBits.One);
                tempPort.Open();

                for (int i = 0; i < 10; i++)
                {
                    tempPort.Write("\" ");
                    Thread.Sleep(100);
                    string readAttempt = tempPort.ReadExisting();
                    if (readAttempt.Split(",")[0] == "PCSELECTOR")
                    {
                        SelectorButtons = int.Parse(readAttempt.Split(",")[1]);
                        ShowSelectorConnected = true;
                        tempPort.Close();
                        ShowSelectorPort = new SerialPort(portDialog.PortName, 9600, Parity.None, 8, StopBits.One);
                        ShowSelectorPort.Open();
                        successPortName = portDialog.PortName;
                        break;
                    }
                }

                tempPort.Close();

                if (!ShowSelectorConnected)
                {
                    MessageBox.Show("Could not detect a PinkConnection Show Selector on the specified serial port.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

            }
            catch (Exception ex)
            {
                MessageBox.Show("Could not detect a PinkConnection Show Selector on the specified serial port.\n" + ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            ShowSelectorGroupBox.Text = $"Show Selector: {SelectorButtons} button on {successPortName}";
            ShowSelectorAutoDetectButton.Enabled = false;
            ShowSelectorSelectPortButton.Enabled = false;
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (Playing) Stop();
            if (InvalidateSave)
                if (MessageBox.Show(
                    "You have an unsaved Playlist!\nDo you still want to exit?",
                    "Unsaved Playlist",
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Exclamation
                    ) != DialogResult.Yes)
                    e.Cancel = true;
        }

        private void SavePlaylistButton_Click(object sender, EventArgs e)
        {
            if (SavePlaylistFileDialog.ShowDialog() == DialogResult.OK)
            {
                File.WriteAllText(SavePlaylistFileDialog.FileName, JsonSerializer.Serialize<PlaylistConfig>(Config));
                InvalidateSave = false;
                PlaylistGroupBox.Text = "Playlist: " + Path.GetFileNameWithoutExtension(SavePlaylistFileDialog.FileName);
            }
        }

        private void LoadPlaylistButton_Click(object sender, EventArgs e)
        {
            if (Playing)
            {
                if (MessageBox.Show(
                    "A show is currently playing!\nPerforming this action will stop it.\nAre you sure?",
                    "Show Playing",
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Exclamation
                    ) != DialogResult.Yes)
                    return;
                Stop();
            }

            if (!ControllerConnected)
            {
                MessageBox.Show("You must connect a Controller first.");
                return;
            }

            if (InvalidateSave)
                if (MessageBox.Show(
                    "You have an unsaved Playlist!\nDo you still want to load another?",
                    "Unsaved Playlist",
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Exclamation
                    ) != DialogResult.Yes)
                    return;

            if (OpenPlaylistFileDialog.ShowDialog() == DialogResult.OK)
            {
                PlaylistConfig? tempConfig;
                try
                {
                    tempConfig = JsonSerializer.Deserialize<PlaylistConfig>(File.ReadAllText(OpenPlaylistFileDialog.FileName));
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Could not read the selected Playlist file.\n" + ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
                if (tempConfig == null)
                {
                    MessageBox.Show("Could not read the selected Playlist file.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                if (tempConfig.Mapping != null)
                {
                    if (tempConfig.Mapping.Bits.Length % 4 != 0)
                    {
                        MessageBox.Show("The mapped channel count is not divisible by 4.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }
                    ChannelMappingGroupBox.Text = $"Channel Map: {tempConfig.Mapping.Name}";
                    ChannelMappingLoaded = true;
                }

                int showIndex = 0;
                foreach (PlaylistShow show in tempConfig.Shows)
                {
                    if (tempConfig.Mapping == null)
                    {
                        MessageBox.Show(
                            "The Playlist does not have a loaded Channel Map" +
                            "\nNo Shows or Skits will be loaded."
                        , "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        break;
                    }
                    if (!File.Exists(show.FilePath))
                    {
                        if (MessageBox.Show(
                            $"Could not locate the file for the Show {show.Name}" +
                            $"\nWould you like to relocate it?",
                            "Could not find Show file",
                            MessageBoxButtons.YesNo,
                            MessageBoxIcon.Exclamation
                            ) == DialogResult.Yes)
                        {
                            if (SelectShowFileDialog.ShowDialog() == DialogResult.OK)
                            {
                                tempConfig.Shows[showIndex].FilePath = SelectShowFileDialog.FileName;
                            }
                            else tempConfig = RemoveShowAt(showIndex, tempConfig);
                        }
                        else tempConfig = RemoveShowAt(showIndex, tempConfig);
                    }
                    if (show.Type != tempConfig.Mapping.Type)
                    {
                        MessageBox.Show(
                            $"Show \"{show.Name}\"'s type does not match the Playlist's Channel Map." +
                            $"\nIt will be removed from the playlist." +
                            $"\nLoaded Channel Map Type: {tempConfig.Mapping.Type}" +
                            $"\nShow Type: {show.Type}"
                        , "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        tempConfig = RemoveShowAt(showIndex, tempConfig);
                    }

                    showIndex++;
                }

                int skitIndex = 0;
                foreach (PlaylistShow skit in tempConfig.Skits)
                {
                    if (tempConfig.Mapping == null) break;
                    if (!File.Exists(skit.FilePath))
                    {
                        if (MessageBox.Show(
                            $"Could not locate the file for the Skit {skit.Name}" +
                            $"\nWould you like to relocate it?",
                            "Could not find Skit file",
                            MessageBoxButtons.YesNo,
                            MessageBoxIcon.Exclamation
                            ) == DialogResult.Yes)
                        {
                            if (SelectShowFileDialog.ShowDialog() == DialogResult.OK)
                            {
                                tempConfig.Skits[skitIndex].FilePath = SelectShowFileDialog.FileName;
                            }
                            else tempConfig.Skits.RemoveAt(skitIndex);
                        }
                        else tempConfig.Skits.RemoveAt(skitIndex);
                    }
                    if (skit.Type != tempConfig.Mapping.Type)
                    {
                        MessageBox.Show(
                            $"Skit \"{skit.Name}\"'s type does not match the Playlist's Channel Map." +
                            $"\nIt will be removed from the playlist." +
                            $"\nLoaded Channel Map Type: {tempConfig.Mapping.Type}" +
                            $"\nSkit Type: {skit.Type}"
                        , "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        tempConfig.Skits.RemoveAt(skitIndex);
                    }

                    skitIndex++;
                }

                Config = tempConfig;
                PlaylistGroupBox.Text = "Playlist: " + Path.GetFileNameWithoutExtension(OpenPlaylistFileDialog.FileName);
                UpdateListView();
                InvalidateSave = false;
            }
        }

        private PlaylistConfig RemoveShowAt(int position, PlaylistConfig pcfg)
        {
            int assindex = 0;
            foreach (int? assignment in pcfg.SelectorAssignments)
            {
                if (assignment == null) continue;
                if (assignment == position) pcfg.SelectorAssignments[assindex] = null;
                if (assignment > position) pcfg.SelectorAssignments[assindex]--;
                assindex++;
            }
            pcfg.Shows.RemoveAt(position);
            return pcfg;
        }

        private void LoadNextShowtape()
        {
            string tempUSTData = File.ReadAllText(CurrentlyPlaying.FilePath);
            if (!tempUSTData.StartsWith("UST,2,"))
            {
                MessageBox.Show($"Selected showtape \"{Path.GetFileNameWithoutExtension(CurrentlyPlaying.FilePath)}\" is not a UST version 2 showtape.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            string[] stringyBits = tempUSTData.Split(';')[1].Split(',');
            if (File.Exists("pc3playertempaudio.tmp")) File.Delete("pc3playertempaudio.tmp");
            File.WriteAllBytes("pc3playertempaudio.tmp", Convert.FromBase64String(tempUSTData.Split(';')[2]));

            tempUSTData = null;

            List<string> tempShowData = new List<string>();

            foreach (string stringyFrame in stringyBits)
            {
                BigInteger frame = BigInteger.Parse(stringyFrame, NumberStyles.HexNumber);
                int selectBit = 0;
                char[] frameStringOut = new char[64];
                for (int i = 0; i < Config.Mapping.Bits.Count() / 4; i++)
                {
                    byte quartet = 64;
                    for (int j = 0; j < 4; j++)
                    {
                        if (Config.Mapping.Bits[selectBit] == 0) continue;
                        if (Config.Mapping.InvertedBits.Contains(Config.Mapping.Bits[selectBit])) { if (!((frame & BigInteger.Pow(2, Config.Mapping.Bits[selectBit] - 1)) == BigInteger.Pow(2, Config.Mapping.Bits[selectBit] - 1))) quartet += (byte)Math.Pow(2, j); }
                        else { if ((frame & BigInteger.Pow(2, Config.Mapping.Bits[selectBit] - 1)) == BigInteger.Pow(2, Config.Mapping.Bits[selectBit] - 1)) quartet += (byte)Math.Pow(2, j); }
                        selectBit++;
                    }
                    frameStringOut[i] = (char)quartet;
                }
                tempShowData.Add(new string(frameStringOut).Trim());
            }

            ShowtapeFrames = tempShowData.ToArray();
        }

        private void SelectNextShowtape(StartMethod method)
        {
            ShowtapeFrames = [];
            ShowtapeIndex = 0;
            if (Config.Shows.Count() == 0)
            {
                MessageBox.Show("There are no shows in the playlist.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }
            bool willStartShow = false;
            bool bypassIntermission = false;
            if (Intermission)
            {
                Intermission = false;
                bypassIntermission = true;
                GUIUpdateTimer.Stop();
            }
            switch (method)
            {
                case StartMethod.PlaylistStart:
                    switch (Config.PlaylistStartOption)
                    {
                        case 0:
                            PlaylistIndex = 0;
                            willStartShow = true;
                            break;
                        case 1:
                            PlaylistIndex = SelectRandom.Next(Config.Shows.Count());
                            willStartShow = true;
                            break;
                    }
                    break;
                case StartMethod.PlaylistShowEnd:
                    if (Config.PlaylistIntermissionOption && (!bypassIntermission))
                    {
                        Intermission = true;
                        IntermissionTotalTime = IntermissionSecondsLeft = Config.PlaylistIntermissionTimeOption * ((Config.PlaylistIntermissionUnitOption == 1) ? 1 : 60);
                        SkipButton.Enabled = true;
                        StopButton.Enabled = true;
                        PauseButton.Enabled = true;
                        Invoke(delegate { PlaybackControlsGroupBox.Text = "Controls: Intermission"; });
                        GUIUpdateTimer.Start();
                    }
                    else
                    {
                        switch (Config.PlaylistShowEndOption)
                        {
                            case 0:
                                PlaylistIndex++;
                                if (PlaylistIndex >= Config.Shows.Count())
                                {
                                    if (Config.PlaylistLoopOption)
                                    {
                                        PlaylistIndex = 0;
                                        willStartShow = true;
                                    }
                                }
                                else willStartShow = true;
                                break;
                            case 1:
                                PlaylistIndex = SelectRandom.Next(Config.Shows.Count());
                                willStartShow = true;
                                break;
                            case 2:
                                willStartShow = true;
                                break;
                            case 3:
                                break;
                        }
                    }
                    break;
            }

            if (willStartShow)
            {
                CurrentlyPlaying = Config.Shows[PlaylistIndex];
                LoadNextShowtape();

                FrameTimer = new System.Timers.Timer((1000d / 60d) * Config.FramesPerTick);
                FrameTimer.Elapsed += PlayFrame;
                FrameTimer.AutoReset = true;

                ResyncTimer = new System.Timers.Timer(Config.ResyncSeconds * 1000);
                ResyncTimer.Elapsed += Resync;
                ResyncTimer.AutoReset = true;

                dataProvider = new StreamDataProvider(File.OpenRead("pc3playertempaudio.tmp"));
                AudioPlayer = new SoundPlayer(dataProvider);

                Invoke(delegate { PlaybackControlsGroupBox.Text = $"Controls: Playing {CurrentlyPlaying.Name}"; });

                Mixer.Master.AddComponent(AudioPlayer);
                Playing = true;
                AudioPlayer.Play();
                FrameTimer.Start();
                ResyncTimer.Start();
                GUIUpdateTimer.Start();
                PauseButton.Enabled = true;
                StopButton.Enabled = true;
                SkipButton.Enabled = true;
            }
        }

        void GUIUpdateTick(object sender, ElapsedEventArgs e)
        {
            if (Playing)
            {
                Invoke(delegate { TimeLabel.Text = $"({TimeSpan.FromSeconds(AudioPlayer.Time).ToString(@"hh\:mm\:ss")}/{CurrentlyPlaying.Length})"; });
            }
            if (Intermission)
            {
                if (IntermissionSecondsLeft == 0) SelectNextShowtape(StartMethod.PlaylistShowEnd);
                Invoke(delegate { TimeLabel.Text = $"({TimeSpan.FromSeconds(IntermissionTotalTime - IntermissionSecondsLeft).ToString(@"hh\:mm\:ss")}/{TimeSpan.FromSeconds(IntermissionTotalTime).ToString(@"hh\:mm\:ss")})"; });
                if (!Paused) IntermissionSecondsLeft--;
            }
        }

        void PlayFrame(object sender, ElapsedEventArgs e)
        {
            if (ShowtapeIndex + Config.FrameShift >= ShowtapeFrames.Length)
            {
                Stop();
                SelectNextShowtape(StartMethod.PlaylistShowEnd);
                return;
            }
            ControllerPort.Write(ShowtapeFrames[ShowtapeIndex + Config.FrameShift]);
            ShowtapeIndex += Config.FramesPerTick;
        }

        void Resync(object sender, ElapsedEventArgs e)
        {
            ShowtapeIndex = (int)(AudioPlayer.Time * 60f);
        }

        private void PauseButton_Click(object sender, EventArgs e)
        {
            if (Paused)
            {
                Paused = false;
                PauseButton.Text = "Pause";
                if (Playing)
                {
                    FrameTimer.Start();
                    ResyncTimer.Start();
                    AudioPlayer.Seek((float)(((float)ShowtapeIndex) / 60.0));
                    AudioPlayer.Play();
                }
            }
            else
            {
                Paused = true;
                PauseButton.Text = "Unpause";
                if (Playing)
                {
                    FrameTimer.Stop();
                    ResyncTimer.Stop();
                    AudioPlayer.Pause();
                }
            }
        }

        private void StopButton_Click(object sender, EventArgs e)
        {
            Stop();
        }

        private void SkipButton_Click(object sender, EventArgs e)
        {
            if (Playing) Stop();
            SelectNextShowtape(StartMethod.PlaylistShowEnd);
        }

        private void StartButton_Click(object sender, EventArgs e)
        {
            if (Playing)
            {
                if (MessageBox.Show(
                    "A show is currently playing!\nPerforming this action will stop it.\nAre you sure?",
                    "Show Playing",
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Exclamation
                    ) != DialogResult.Yes)
                    return;
                Stop();
            }
            SelectNextShowtape(StartMethod.PlaylistStart);
        }

        void Stop()
        {
            Playing = false;
            AudioPlayer.Stop();
            FrameTimer.Stop();
            ResyncTimer.Stop();
            GUIUpdateTimer.Stop();
            ResyncTimer.Dispose();
            FrameTimer.Dispose();
            PauseButton.Enabled = false;
            StopButton.Enabled = false;
            SkipButton.Enabled = false;
            if (Playing) Mixer.Master.RemoveComponent(AudioPlayer);
            Paused = false;
            Invoke(delegate
            {
                PlaybackControlsGroupBox.Text = "Controls: Not Playing";
                PauseButton.Text = "Pause";
                TimeLabel.Text = "(00:00:00/00:00:00)";
            });
            dataProvider.Dispose();
            if (File.Exists("pc3playertempaudio.tmp")) File.Delete("pc3playertempaudio.tmp");
            ShowtapeFrames = [];
            ShowtapeIndex = 0;
        }

        private void NewPlaylistButton_Click(object sender, EventArgs e)
        {
            if (Playing)
            {
                if (MessageBox.Show(
                    "A show is currently playing!\nPerforming this action will stop it.\nAre you sure?",
                    "Show Playing",
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Exclamation
                    ) != DialogResult.Yes)
                    return;
                Stop();
            }

            if (InvalidateSave)
                if (MessageBox.Show(
                    "You have an unsaved Playlist!\nDo you still want to create a new one?",
                    "Unsaved Playlist",
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Exclamation
                    ) != DialogResult.Yes)
                    return;

            Config = new PlaylistConfig();
            ChannelMappingGroupBox.Text = "Channel Map: None Loaded";
            PlaylistGroupBox.Text = "Playlist: Unsaved";
            UpdateListView();
            InvalidateSave = false;
        }
    }
}
